package info.textgrid.middleware;



import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;

import java.math.BigInteger;
import java.text.ParseException;

import javax.xml.datatype.DatatypeConfigurationException;

import org.apache.commons.logging.LogFactory;
import org.elasticsearch.action.search.SearchRequestBuilder;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.search.SearchType;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;



/**
 * Class to build the Element for a ListIdentifiers request
 * 
 * @author Maximilian Brodhun, SUB
 * @version 1.0
 * @since 20.02.2014
 */

public class IdentifierListDeliverer {

	private static org.apache.commons.logging.Log	log		= LogFactory
																	.getLog(IdentifierListDeliverer.class);
	private String									datestamp;
	private boolean									idExist	= true;
	private OAI_ESClient							oaiEsClient;

	private boolean									textgrid;
	private boolean									dariah;

	private String[]								identifierListFields;  //Fields for the elasticsearch request
	private String									rangeField; 		// Field for the optional range query 
	private String									formatToFilter;		// just necessary in textgrid case to just get the editions
	private String									formatField;		// just necessary in textgrid to just get the editions
	private String									dateOfObjectCreation; // Field for the object creation in the repository 
	private String									repositoryObjectURIPrefix; //
	private String									identifierField;
	long start = 0;
	int searchResponseSize = 5;
	String resumptionToken;
	ResumptionTokenType value = new ResumptionTokenType();
	
	//private String 									set;

	/**
	 * In OAIPMH a ListIdentifiers request is answered by responding the
	 * datestamp and the identifier of an object Both will be responded in a
	 * header element. This function build the header element
	 * 
	 * @param datestamp
	 *            taken from the elasticSearch index (created element)
	 * @param identifier
	 *            taken from the elasticSearch index (textgridUri)
	 * @return the ListIdentifierElement with the responded header element
	 */

	public IdentifierListDeliverer(OAI_ESClient oaiEsClient, boolean textgrid,
			boolean dariah) {
		this.textgrid = textgrid;
		this.dariah = dariah;
		this.oaiEsClient = oaiEsClient;
		
	}
	
	
	// FIXME Delete this function and check for empty set string in setListIdentifierHeaderWithSet
	
	private ListIdentifiersType setListIdentifierHeader(String datestamp,
			String identifier, ListIdentifiersType lit, String set) {
		
		HeaderType header = new HeaderType();
		
		header.setDatestamp(datestamp);
		header.setIdentifier(identifier);
		
		if(set!=null){
			header.getSetSpec().add(set);
		}
		
		lit.getHeader().add(header);

		return lit;
	}

	/**
	 * To get the required values for the ListIdentifiers request this function
	 * will ask ElasticSearch for a specific textgridUri the values "created"
	 * and "textgridUri".
	 * 
	 * Since the ListIdentifiers request enables the possibility the limit the
	 * request in a specific time interval, it is necessary to perform a range
	 * query on the ElasticSearch index
	 * 
	 * @param from
	 *            : start value for the range query
	 * @param to
	 *            : end value to the range query
	 * @return after calling the function "setListIdentifierHeader" the return
	 *         value is the whole ListIdentifiers element
	 * @throws ParseException
	 */
	
	// FIXME combine it with with getIdentifierListWithSet and check for empty set
	
	
	public ListIdentifiersType processIdentifierList(String from, String to, String set)
			throws ParseException {
		ListIdentifiersType lit = new ListIdentifiersType();
		
		String identifier = "";
		QueryBuilder query =  setOrNot(set, from, to);
			
		SearchRequestBuilder getListIdentifiersValuesBuilder = oaiEsClient
				.getOaiESClient().prepareSearch(oaiEsClient.getEsIndex())
				.setSearchType(SearchType.SCAN)				
				.setScroll(new TimeValue(60000))
				.setTypes(oaiEsClient.getEsType())
				.addFields(identifierListFields)
				.setQuery(query)				
				.setSize(searchResponseSize);
		
		SearchResponse listListIdentiferValues = getListIdentifiersValuesBuilder.execute().actionGet();		
		
		BigInteger bi2 = BigInteger.valueOf(listListIdentiferValues.getHits().getTotalHits()); //Number of Total Hits as BigInteger (required by oaipmh ResumptionToken)
		BigInteger bi1 = BigInteger.valueOf(start); //start Value of the list. Important to increase the the cursor value of the resumption token
		
		//if (listListIdentiferValues.getHits().getTotalHits() > 0) {
		while (bi1.floatValue() < bi2.floatValue()) {
			listListIdentiferValues =hitHandling(listListIdentiferValues,  listListIdentiferValues.getScrollId(), identifier,  lit,  set);
			
			listListIdentiferValues = oaiEsClient.getOaiESClient()
					.prepareSearchScroll(listListIdentiferValues.getScrollId())
					.setScroll(new TimeValue(60000))
					.execute()
					.actionGet();			
			
			start=start+searchResponseSize;					
			
			
			lit.setResumptionToken(setResToken(bi2, bi1, listListIdentiferValues.getScrollId()));
			
			if (listListIdentiferValues.getHits().getHits().length == 0) {
		        break;
		    }
			
		}

		return lit;
	}

	public SearchResponse hitHandling(SearchResponse listFurtherValues, String scrollID, String identifier, ListIdentifiersType lit, String set){
		for (SearchHit hit : listFurtherValues.getHits().getHits()) {
			
			if (this.textgrid && hit.getFields().get(formatField).getValue().toString().equals(formatToFilter)) {
				
				try {
					String datestamp = hit.getFields().get(this.dateOfObjectCreation).getValue().toString();
					datestamp = OAIPMHUtillities.convertDateFormat(datestamp).toXMLFormat();

				} catch (ParseException e1) {
					log.debug(e1);
				} catch (DatatypeConfigurationException e1) {
					log.debug(e1);
				}	
				identifier = hit.getFields().get(identifierField).getValue().toString();
				lit = setListIdentifierHeader(datestamp, identifier, lit, set);
			}				

			if (this.dariah) {					
				datestamp = hit.getFields().get(dateOfObjectCreation).getValue().toString();

				try {
					String datestamp = hit.getFields().get(this.dateOfObjectCreation).getValue().toString();
					datestamp = OAIPMHUtillities.convertDateFormat(datestamp).toXMLFormat();
				} catch (ParseException e1) {
					log.debug(e1);
				} catch (DatatypeConfigurationException e1) {
					log.debug(e1);
				}
				
				identifier = hit.getFields().get(identifierField).getValue().toString();					
				lit = setListIdentifierHeader(datestamp, identifier, lit, set);
			}
		}
		return listFurtherValues;
	}
	
	public QueryBuilder setOrNot(String set, String from, String to){
		
		QueryBuilder rangeQuery = QueryBuilders.rangeQuery(rangeField).from(from).to(to);
		QueryBuilder formatQuery = QueryBuilders.matchPhrasePrefixQuery("format", "text/tg.edition+tg.aggregation+xml");
		QueryBuilder query;
		
		if(set!=null){
			String queryField = "";
			String valueField = "";
			
			if (this.dariah == true) {
				/*
				 * COLLECTIONREGISTRY_NAMESPACE is the name of the set Category
				 * but no field in ElasticSearch. It has to be queried in field
				 * "metadata.dc:relation".
				 */				
				queryField = "metadata.dc:relation";
				valueField = set;		
			}

			if (this.textgrid == true) {
				String[] setParts = set.split(":");
				if (setParts[0].equals("project")) {
					queryField = "project.id";
					valueField = setParts[1];
				}
			}
				
			QueryBuilder matchQuery = QueryBuilders.matchPhraseQuery(queryField, valueField);
			
			query = QueryBuilders.boolQuery().must(matchQuery).must(formatQuery);
						
		} else {
			query = QueryBuilders.boolQuery().must(rangeQuery).must(formatQuery);
		}
		
		return query;		
	}
	
	public ResumptionTokenType setResToken(BigInteger bi2, BigInteger bi1, String scrollID){
		
		value.setCompleteListSize(bi2);
		value.setCursor(bi1);
		resumptionToken = scrollID;
		value.setValue(resumptionToken);
		
		return value;
	}
	
	public ListIdentifiersType getLit(ListIdentifiersType lit) {
		return lit;
	}

	/**
	 * checking the request if all necessary values are set or not allowed
	 * values are not set
	 * 
	 * @param request
	 *            the initially request from the OAIPMG client
	 * @return a boolean indicating the correctness of the request
	 */

	public boolean requestChecker(RequestType request) {
		boolean requestCheck;

		if (request.getMetadataPrefix() == null
				|| request.getIdentifier() != null) {

			requestCheck = false;
		} else {
			requestCheck = true;
		}

		return requestCheck;

	}

	public void setFormatToFilter(String formatToFilter) {
		this.formatToFilter = formatToFilter;
	}

	public void setFormatField(String formatField) {
		this.formatField = formatField;
	}

	public String getDateOfObjectCreation() {
		return dateOfObjectCreation;
	}

	public void setDateOfObjectCreation(String dateOfObjectCreation) {
		this.dateOfObjectCreation = dateOfObjectCreation;
	}

	public void setRepositoryObjectURIPrefix(String repositoryObjectURIPrefix) {
		this.repositoryObjectURIPrefix = repositoryObjectURIPrefix;
	}

	public String getRepositoryObjectURIPrefix() {
		return repositoryObjectURIPrefix;
	}

	public void setIdExist(boolean validItem) {
		idExist = validItem;
	}

	public boolean getIdExist() {
		return idExist;
	}

	public String[] getIdentifierListFields() {
		return identifierListFields;
	}

	public void setIdentifierListFields(String[] identifierListFields) {
		this.identifierListFields = identifierListFields;
	}

	public String getFieldForRange() {
		return rangeField;
	}

	public void setFieldForRange(String fieldForRange) {
		this.rangeField = fieldForRange;
	}
	
	public String getIdentifierField() {
		return identifierField;
	}

	public void setIdentifierField(String identifierPrefix) {
		this.identifierField = identifierPrefix;
	}
}
